use std::io::BufRead;
use std::io::Write;

const LABYRINTH_COUNT: usize = 7;

fn prompt_and_read(prompt: &str) -> String {
    let mut line = String::new();
    let stdin = std::io::stdin();
    let mut stdout = std::io::stdout();
    print!("{}", prompt);

    stdout.flush().unwrap();
    stdin.lock().read_line(&mut line).unwrap();
    line
}

fn get_labyrinth_index() -> usize {
    let mut i = None;
    loop {
        let line = prompt_and_read("Give me an index: ");
        if let Ok(index) = line.trim_end().parse::<usize>() {
            i = Some(index);
            if index >= LABYRINTH_COUNT {
                println!("Please give me a number less than {}", LABYRINTH_COUNT);
                continue;
            }
        } else {
            println!("Please give me a number");
        }

        if let Some(i) = i {
            break i;
        }
    }
}

fn solve_labyrinth(labyrinth: &mut u64) {
    println!("The labyrinth consists of 32 steps; in order to solve it, you have to guess each step's direction correctly\n");
    println!("Go east: l");
    println!("Go south: j");
    println!("Go west: h");
    println!("Go north: k");
    let mut i = 0;
    while i < 32 {
        let line = prompt_and_read(&format!("Give me the {}. direction: ", i));
        if let Some(x) = match line.chars().next().unwrap() {
            'l' => Some(0b00),
            'j' => Some(0b01),
            'h' => Some(0b10),
            'k' => Some(0b11),
            _ => None,
        } {
            let shift = 2 * i;
            let value = x << shift;
            *labyrinth ^= value;
            i += 1;
        } else {
            println!("Please give me a direction (h, j, k or l)");
        }
    }
}

fn check_solution(labyrinths: &[u64]) -> bool {
    labyrinths.iter().all(|labyrinth| *labyrinth == 0)
}

fn salvation() {
    println!(
        "{}",
        String::from_utf8(std::fs::read("flag.txt").unwrap()).unwrap()
    );
}

fn challenge() {
    let mut labyrinths: [u64; LABYRINTH_COUNT] = [0; LABYRINTH_COUNT];
    labyrinths.iter_mut().for_each(|lab| *lab = rand::random());
    // println!("{:#x?}", labyrinths);
    loop {
        println!("Please make a choice: ");
        println!("1: Solve a labyrinth");
        println!("2: Check solution");
        let line = prompt_and_read("Choice: ");
        match line.trim().parse::<u8>() {
            Ok(1) => {
                let lab_index = get_labyrinth_index();
                // We already checked that the index is inbounds in `get_labyrinth_index`
                // We gain some performance if we omit the bounds check here
                let labyrinth = unsafe { labyrinths.get_unchecked_mut(lab_index) };
                solve_labyrinth(labyrinth);
            }
            Ok(2) => break,
            _ => println!("Invalid choice"),
        }
    }
    if check_solution(&labyrinths) {
        salvation()
    } else {
        println!("WRONG");
    }
}

fn main() {
    println!("There is no rest for the wicked");
    println!("To reach salvation, you need to pass through {} labyrinths. Oh, and you need to do it blindly.", LABYRINTH_COUNT);
    challenge();
}
